# Examples

```@setup ncaexample
ENV["GKSwstype"] = "nul"
```

## Import

Use [`pkimport`](@ref) to import PK data from table to subject set.

```@example ncaexample
using MetidaNCA, CSV, DataFrames;

pkdata2 = CSV.File(joinpath(dirname(pathof(MetidaNCA)), "..", "test", "csv",  "pkdata2.csv")) |> DataFrame

ds = pkimport(pkdata2, :Time, :Concentration, [:Subject, :Formulation]; dosetime = DoseTime(dose = 100, time = 0))

sort!(ds, :Subject)
```

## NCA

Perform NCA analysis with [`nca!`](@ref). Access to result set is similar to DataFrame or any table.
Find parameter list [here](@ref parameter_list).

```@example ncaexample
dsnca = nca!(ds, adm = :ev, calcm = :lint)

dsnca[:, :AUClast]
```


## Print output

```@example ncaexample
dsnca = nca!(ds[1], adm = :ev, calcm = :lint, verbose = 2);

```

## Plotting

```@example ncaexample
using Plots

p = pkplot(ds; typesort = :Subject, pagesort = nothing, sort = Dict(:Formulation => "R"))

png(p, "plot1.png")

p = pkplot(ds; typesort = :Formulation, pagesort = nothing, legend = true)

png(p, "plot2.png")

p = pkplot(ds; elim = true, ls = true)

png(p[1], "plot3.png")

p = pkplot(ds; typesort = :Subject, pagesort = :Formulation)

png(p[1], "plot4.png")
```

#### Plot 1

![](plot1.png)

#### Plot 2

![](plot2.png)

#### Plot 3

![](plot3.png)

#### Plot 4

![](plot4.png)

## Set dose time

You can set dose time with [`setdosetime!`](@ref) for whole subject set or for
selected subjects.

```@example ncaexample
dt = DoseTime(dose = 200, time = 0)

setdosetime!(ds, dt, Dict(:Formulation => "R"))

dsnca = nca!(ds)

dsnca[:, :Dose]
```

## Set range for elimination

By default no exclusion or range specified. With [`setkelrange!`](@ref) elimination range and exclusion
can be specified for whole subject set or for any selected subjects.

```@example ncaexample
kr =  ElimRange(kelstart = 4, kelend = 12, kelexcl = Int[5,6])

setkelrange!(ds, kr, [1,2,3])

dsnca = nca!(ds)

p = pkplot(ds[1]; elim = true)

png(p, "plot5.png")

getkeldata(ds[1])
```

#### Plot 5

![](plot5.png)


## Without import

You  can use [`nca`](@ref) for NCA analysis directly from tabular data.

```@example ncaexample

dsnca = nca(pkdata2, :Time, :Concentration, [:Subject, :Formulation]; dosetime = DoseTime(dose = 100, time = 0))

sort!(dsnca, :Subject)

dsnca[:, :AUClast]
```
